# Tensor


def cp_decompose(
    self,
    rank,
    bond_ind=None,
    **kwargs,
):
    """Construct a CANDECOMP/PARAFAC decomposition of this tensor via
    fitting.
    """
    t0 = self.copy()

    # adding some noise to the initial tensor (not just factors) makes this
    # much more stable when very symmetric XXX: use fixed seed?
    noise_scale = self.largest_element() * 0.01
    noise = noise_scale * do("random.normal", size=self.shape, like=t0.data)
    t0.modify(data=t0.data + noise)

    if bond_ind is None:
        bond_ind = rand_uuid()

    # construct out initial guess
    tn = TensorNetwork()
    for ix in t0.inds:
        tn |= t0.split(
            left_inds=[ix],
            bond_ind=bond_ind,
            max_bond=rank,
            cutoff=0.0,
            get="tensors",
        )[0]

    # fit and return the new TN
    return tn.fit_(self, **kwargs)


# TensorNetwork


def cp_simplify(
    self,
    atol=1e-12,
    equalize_norms=False,
    cache=None,
    inplace=False,
    **cp_opts,
):
    tn = self if inplace else self.copy()

    # we don't want to repeatedly check the CP decompositions of the
    #     same tensor as we cycle through simplification methods
    if cache is None:
        cache = set()

    for tid, t in tuple(tn.tensor_map.items()):
        # id's are reused when objects go out of scope -> use tid as well
        cache_key = ("cp", tid, id(t.data))
        if cache_key in cache:
            continue

        if t.ndim < 3:
            cache.add(cache_key)
            continue

        try:
            # all dimensions need to be the same
            (rank,) = set(t.shape)
        except ValueError:
            cache.add(cache_key)
            continue

        # compute the decomposition
        tcp = t.cp_decompose(rank, **cp_opts)

        if any(np.any(~np.isfinite(t.data)) for t in tcp):
            raise ValueError("bad value")

        # check if its exact up to atol
        err = tcp.distance(t)
        if err < atol:
            if equalize_norms:
                tcp.equalize_norms(value=equalize_norms)

            tn.pop_tensor(tid)
            tn |= tcp
        else:
            cache.add(cache_key)

    return tn


cp_simplify_ = functools.partialmethod(cp_simplify, inplace=True)



import warnings

from ...operator.builder import (
    SparseOperatorBuilder,
    get_mat,
)
from ...operator.hilbertspace import (
    HilbertSpace,
)
from ...operator.models import (
    fermi_hubbard_from_edges,
    fermi_hubbard_spinless_from_edges,
    heisenberg_from_edges,
    rand_operator,
)

__all__ = (
    "fermi_hubbard_from_edges",
    "fermi_hubbard_spinless_from_edges",
    "get_mat",
    "heisenberg_from_edges",
    "HilbertSpace",
    "rand_operator",
    "SparseOperatorBuilder",
)


warnings.warn(
    "Most functionality of 'quimb.experimental.operatorbuilder' "
    "has been moved to `quimb.operator`.",
)



import pytest
from numpy.testing import assert_allclose

import quimb as qu
import quimb.operator as qop
import quimb.tensor as qtn


@pytest.mark.parametrize(
    "sector,size",
    [
        (None, 4096),
        ("even", 2048),
        (6, 924),
    ],
)
def test_heisenberg_square(sector, size):
    edges = (
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2)),
        ((0, 3), (1, 3)),
        ((1, 0), (1, 1)),
        ((1, 0), (2, 0)),
        ((1, 1), (1, 2)),
        ((1, 1), (2, 1)),
        ((1, 2), (1, 3)),
        ((1, 2), (2, 2)),
        ((1, 3), (2, 3)),
        ((2, 0), (2, 1)),
        ((2, 1), (2, 2)),
        ((2, 2), (2, 3)),
    )
    sop = qop.heisenberg_from_edges(edges)
    assert sop.hilbert_space.get_size(sector) == size
    H = sop.build_sparse_matrix(sector=sector)
    assert H.shape == (size, size)
    assert qu.groundenergy(H) == pytest.approx(-6.69168019351495)


def test_heisenberg_mpo():
    L = 8
    Href = qu.ham_heis(L)
    edges = qtn.edges_1d_chain(L)
    sop = qop.heisenberg_from_edges(edges)

    mpo = sop.build_mpo()
    assert mpo.nsites == L
    assert mpo.max_bond() == 5
    assert_allclose(mpo.to_dense(), Href, atol=1e-10)

    sop.pauli_decompose()
    mpo = sop.build_mpo()
    assert mpo.nsites == L
    assert mpo.max_bond() == 5
    assert_allclose(mpo.to_dense(), Href, atol=1e-10)

    sop.pauli_decompose(use_zx=True)
    mpo = sop.build_mpo()
    assert mpo.nsites == L
    assert mpo.max_bond() == 5
    assert_allclose(mpo.to_dense(), Href, atol=1e-10)


def test_fermi_hubbard_hex():
    # quspin reference:
    # https://quspin.github.io/QuSpin/examples/example18.html#example18-label
    # hexagonal graph
    edges = [
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (0, 2)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2)),
        ((0, 3), (0, 4)),
        ((0, 4), (1, 4)),
        ((1, 0), (1, 1)),
        ((1, 1), (1, 2)),
        ((1, 1), (2, 1)),
        ((1, 2), (1, 3)),
        ((1, 3), (1, 4)),
        ((1, 3), (2, 3)),
        ((1, 4), (1, 5)),
        ((1, 5), (2, 5)),
        ((2, 1), (2, 2)),
        ((2, 2), (2, 3)),
        ((2, 3), (2, 4)),
        ((2, 4), (2, 5)),
    ]
    N_up = 2  # number of spin-up fermions
    N_down = 2  # number of spin-down fermions
    t = 1.0  # tunnelling matrix element
    U = 2.0  # on-site fermion interaction strength
    sob = qop.fermi_hubbard_from_edges(edges, t=t, U=U)
    sector = (
        (sob.nsites // 2, N_up),
        (sob.nsites // 2, N_down),
    )
    assert sob.hilbert_space.get_size(sector) == 14400
    # build sparse matrix
    H = sob.build_sparse_matrix(sector=sector)
    # solve for groundstate
    energy = qu.groundenergy(H)
    assert energy == pytest.approx(-8.67415949)



"""Functions relating to tensor network contraction."""

import collections
import contextlib
import functools
import itertools
import threading

import cotengra as ctg

_CONTRACT_STRATEGY = "greedy"
_TEMP_CONTRACT_STRATEGIES = collections.defaultdict(list)


def get_contract_strategy():
    r"""Get the default contraction strategy - the option supplied as
    ``optimize`` to ``cotengra``.
    """
    if not _TEMP_CONTRACT_STRATEGIES:
        # shortcut for when no temp strategies are in use
        return _CONTRACT_STRATEGY

    thread_id = threading.get_ident()
    if thread_id not in _TEMP_CONTRACT_STRATEGIES:
        return _CONTRACT_STRATEGY

    temp_strategies = _TEMP_CONTRACT_STRATEGIES[thread_id]
    # empty list -> not in context manager -> use default strategy
    if not temp_strategies:
        # clean up to allow above shortcuts
        del _TEMP_CONTRACT_STRATEGIES[thread_id]
        return _CONTRACT_STRATEGY

    # use most recently set strategy for this threy
    return temp_strategies[-1]


def set_contract_strategy(strategy):
    """Get the default contraction strategy - the option supplied as
    ``optimize`` to ``cotengra``.
    """
    global _CONTRACT_STRATEGY
    _CONTRACT_STRATEGY = strategy


@contextlib.contextmanager
def contract_strategy(strategy, set_globally=False):
    """A context manager to temporarily set the default contraction strategy
    supplied as ``optimize`` to ``cotengra``. By default, this only sets the
    contract strategy for the current thread.

    Parameters
    ----------
    set_globally : bool, optimize
        Whether to set the strategy just for this thread, or for all threads.
        If you are entering the context, *then* using multithreading, you might
        want ``True``.
    """
    if set_globally:
        orig_strategy = get_contract_strategy()
        set_contract_strategy(strategy)
        try:
            yield
        finally:
            set_contract_strategy(orig_strategy)
    else:
        thread_id = threading.get_ident()
        temp_strategies = _TEMP_CONTRACT_STRATEGIES[thread_id]
        temp_strategies.append(strategy)
        try:
            yield
        finally:
            temp_strategies.pop()


get_symbol = ctg.get_symbol


def empty_symbol_map():
    """Get a default dictionary that will populate with symbol entries as they
    are accessed.
    """
    return collections.defaultdict(map(get_symbol, itertools.count()).__next__)


def inds_to_symbols(inputs):
    """Map a sequence of inputs terms, containing any hashable indices, to
    single unicode letters, appropriate for einsum.

    Parameters
    ----------
    inputs : sequence of sequence of hashable
        The input indices per tensor.

    Returns
    -------
    symbols : dict[hashable, str]
        The mapping from index to symbol.
    """
    return ctg.get_symbol_map(inputs)


@functools.lru_cache(2**12)
def inds_to_eq(inputs, output=None):
    """Turn input and output indices of any sort into a single 'equation'
    string where each index is a single 'symbol' (unicode character).

    Parameters
    ----------
    inputs : sequence of sequence of hashable
        The input indices per tensor.
    output : sequence of hashable
        The output indices.

    Returns
    -------
    eq : str
        The string to feed to einsum/contract.
    """
    symbols = empty_symbol_map()
    in_str = ("".join(symbols[ix] for ix in inds) for inds in inputs)
    in_str = ",".join(in_str)
    if output is None:
        out_str = "".join(
            ix for ix in symbols.values() if in_str.count(ix) == 1
        )
    else:
        out_str = "".join(symbols[ix] for ix in output)
    return f"{in_str}->{out_str}"


_CONTRACT_BACKEND = None
_TENSOR_LINOP_BACKEND = None
_TEMP_CONTRACT_BACKENDS = collections.defaultdict(list)
_TEMP_TENSOR_LINOP_BACKENDS = collections.defaultdict(list)


def get_contract_backend():
    """Get the default backend used for tensor contractions, via 'cotengra'.

    See Also
    --------
    set_contract_backend, get_tensor_linop_backend, set_tensor_linop_backend,
    tensor_contract
    """
    if not _TEMP_CONTRACT_BACKENDS:
        return _CONTRACT_BACKEND

    thread_id = threading.get_ident()
    if thread_id not in _TEMP_CONTRACT_BACKENDS:
        return _CONTRACT_BACKEND

    temp_backends = _TEMP_CONTRACT_BACKENDS[thread_id]
    if not temp_backends:
        del _TEMP_CONTRACT_BACKENDS[thread_id]
        return _CONTRACT_BACKEND

    return temp_backends[-1]


def set_contract_backend(backend):
    """Set the default backend used for tensor contractions, via 'cotengra'.

    See Also
    --------
    get_contract_backend, set_tensor_linop_backend, get_tensor_linop_backend,
    tensor_contract
    """
    global _CONTRACT_BACKEND
    _CONTRACT_BACKEND = backend


@contextlib.contextmanager
def contract_backend(backend, set_globally=False):
    """A context manager to temporarily set the default backend used for tensor
    contractions, via 'cotengra'. By default, this only sets the contract
    backend for the current thread.

    Parameters
    ----------
    set_globally : bool, optimize
        Whether to set the backend just for this thread, or for all threads. If
        you are entering the context, *then* using multithreading, you might
        want ``True``.
    """
    if set_globally:
        orig_backend = get_contract_backend()
        set_contract_backend(backend)
        try:
            yield
        finally:
            set_contract_backend(orig_backend)
    else:
        thread_id = threading.get_ident()
        temp_backends = _TEMP_CONTRACT_BACKENDS[thread_id]
        temp_backends.append(backend)
        try:
            yield
        finally:
            temp_backends.pop()


def get_tensor_linop_backend():
    """Get the default backend used for tensor network linear operators, via
    'cotengra'. This is different from the default contraction backend as
    the contractions are likely repeatedly called many times.

    See Also
    --------
    set_tensor_linop_backend, set_contract_backend, get_contract_backend,
    TNLinearOperator
    """
    if not _TEMP_TENSOR_LINOP_BACKENDS:
        return _TENSOR_LINOP_BACKEND

    thread_id = threading.get_ident()
    if thread_id not in _TEMP_TENSOR_LINOP_BACKENDS:
        return _TENSOR_LINOP_BACKEND

    temp_backends = _TEMP_TENSOR_LINOP_BACKENDS[thread_id]
    if not temp_backends:
        del _TEMP_TENSOR_LINOP_BACKENDS[thread_id]
        return _TENSOR_LINOP_BACKEND

    return temp_backends[-1]


def set_tensor_linop_backend(backend):
    """Set the default backend used for tensor network linear operators, via
    'cotengra'. This is different from the default contraction backend as
    the contractions are likely repeatedly called many times.

    See Also
    --------
    get_tensor_linop_backend, set_contract_backend, get_contract_backend,
    TNLinearOperator
    """
    global _TENSOR_LINOP_BACKEND
    _TENSOR_LINOP_BACKEND = backend


@contextlib.contextmanager
def tensor_linop_backend(backend, set_globally=False):
    """A context manager to temporarily set the default backend used for tensor
    network linear operators, via 'cotengra'. By default, this
    only sets the contract backend for the current thread.

    Parameters
    ----------
    set_globally : bool, optimize
        Whether to set the backend just for this thread, or for all threads. If
        you are entering the context, *then* using multithreading, you might
        want ``True``.
    """
    if set_globally:
        orig_backend = get_tensor_linop_backend()
        set_tensor_linop_backend(backend)
        try:
            yield
        finally:
            set_tensor_linop_backend(orig_backend)
    else:
        thread_id = threading.get_ident()
        temp_backends = _TEMP_TENSOR_LINOP_BACKENDS[thread_id]
        temp_backends.append(backend)
        try:
            yield
        finally:
            temp_backends.pop()


@functools.wraps(ctg.array_contract)
def array_contract(
    arrays,
    inputs,
    output=None,
    optimize=None,
    backend=None,
    **kwargs,
):
    if optimize is None:
        optimize = get_contract_strategy()
    if backend is None:
        backend = get_contract_backend()
    return ctg.array_contract(
        arrays,
        inputs,
        output,
        optimize=optimize,
        backend=backend,
        **kwargs,
    )


@functools.wraps(ctg.array_contract_expression)
def array_contract_expression(*args, optimize=None, **kwargs):
    if optimize is None:
        optimize = get_contract_strategy()
    return ctg.array_contract_expression(*args, optimize=optimize, **kwargs)


@functools.wraps(ctg.array_contract_tree)
def array_contract_tree(*args, optimize=None, **kwargs):
    if optimize is None:
        optimize = get_contract_strategy()
    return ctg.array_contract_tree(*args, optimize=optimize, **kwargs)


@functools.wraps(ctg.array_contract_path)
def array_contract_path(*args, optimize=None, **kwargs):
    if optimize is None:
        optimize = get_contract_strategy()
    return ctg.array_contract_path(*args, optimize=optimize, **kwargs)


def array_contract_pathinfo(*args, **kwargs):
    import opt_einsum as oe

    tree = array_contract_tree(*args, **kwargs)

    if tree.sliced_inds:
        import warnings

        warnings.warn(
            "The contraction tree has sliced indices, which are not "
            "supported by opt_einsum. Ignoring them for now."
        )

    shapes = tree.get_shapes()
    path = tree.get_path()
    eq = tree.get_eq()

    if (eq == "->") and (len(path) == 0):
        # XXX: opt_einsum does not support empty paths
        # https://github.com/jcmgray/quimb/issues/231
        # https://github.com/dgasmith/opt_einsum/pull/229
        path = ((0,),)

    return oe.contract_path(eq, *shapes, shapes=True, optimize=path)[1]



import warnings

from quimb.tensor.belief_propagation import *

warnings.warn(
    "Most functionality of 'quimb.experimental.belief_propagation' "
    "has been moved to `quimb.tensor.belief_propagation`.",
)



"""Hyper, vectorized, 1-norm, belief propagation."""

import autoray as ar

from quimb.tensor.contraction import array_contract
from .bp_common import (
    BeliefPropagationCommon,
    compute_all_index_marginals_from_messages,
    contract_hyper_messages,
    initialize_hyper_messages,
    maybe_get_thread_pool,
)


def _compute_all_hyperind_messages_tree_batched(bm):
    """ """
    ndim = len(bm)

    if ndim == 2:
        # shortcut for 'bonds', which just swap places
        return ar.do("flip", bm, (0,))

    backend = ar.infer_backend(bm)
    _prod = ar.get_lib_fn(backend, "prod")
    _empty_like = ar.get_lib_fn(backend, "empty_like")

    bmo = _empty_like(bm)
    queue = [(tuple(range(ndim)), 1, bm)]

    while queue:
        js, x, bm = queue.pop()

        ndim = len(bm)
        if ndim == 1:
            # reached single message
            bmo[js[0]] = x
            continue
        elif ndim == 2:
            # shortcut for 2 messages left
            bmo[js[0]] = x * bm[1]
            bmo[js[1]] = bm[0] * x
            continue

        # else split in two and contract each half
        k = ndim // 2
        jl, jr = js[:k], js[k:]
        bml, bmr = bm[:k], bm[k:]

        # contract the right messages to get new left array
        xl = x * _prod(bmr, axis=0)

        # contract the left messages to get new right array
        xr = _prod(bml, axis=0) * x

        # add the queue for possible further halving
        queue.append((jl, xl, bml))
        queue.append((jr, xr, bmr))

    return bmo


def _compute_all_hyperind_messages_prod_batched(bm, smudge_factor=1e-12):
    """ """
    backend = ar.infer_backend(bm)
    _prod = ar.get_lib_fn(backend, "prod")
    _reshape = ar.get_lib_fn(backend, "reshape")

    ndim = len(bm)
    if ndim == 2:
        # shortcut for 'bonds', which just swap
        return ar.do("flip", bm, (0,))

    combined = _prod(bm, axis=0)
    return _reshape(combined, (1, *ar.shape(combined))) / (bm + smudge_factor)


def _compute_all_tensor_messages_tree_batched(bx, bm):
    """Compute all output messages for a stacked tensor and messages."""
    backend = ar.infer_backend_multi(bx, bm)
    _stack = ar.get_lib_fn(backend, "stack")

    ndim = len(bm)
    mouts = [None for _ in range(ndim)]
    queue = [(tuple(range(ndim)), bx, bm)]

    while queue:
        js, bx, bm = queue.pop()

        ndim = len(bm)
        if ndim == 1:
            # reached single message
            mouts[js[0]] = bx
            continue
        elif ndim == 2:
            # shortcut for 2 messages left
            mouts[js[0]] = array_contract(
                arrays=(bx, bm[1]),
                inputs=(("X", "a", "b"), ("X", "b")),
                output=("X", "a"),
                backend=backend,
            )
            mouts[js[1]] = array_contract(
                arrays=(bm[0], bx),
                inputs=(("X", "a"), ("X", "a", "b")),
                output=("X", "b"),
                backend=backend,
            )
            continue

        # else split in two and contract each half
        k = ndim // 2
        jl, jr = js[:k], js[k:]
        ml, mr = bm[:k], bm[k:]

        # contract the right messages to get new left array
        xl = array_contract(
            arrays=(bx, *(mr[p] for p in range(mr.shape[0]))),
            inputs=((-1, *js), *((-1, j) for j in jr)),
            output=(-1, *jl),
            backend=backend,
        )

        # contract the left messages to get new right array
        xr = array_contract(
            arrays=(bx, *(ml[p] for p in range(ml.shape[0]))),
            inputs=((-1, *js), *((-1, j) for j in jl)),
            output=(-1, *jr),
            backend=backend,
        )

        # add the queue for possible further halving
        queue.append((jl, xl, ml))
        queue.append((jr, xr, mr))

    return _stack(tuple(mouts))


def _compute_all_tensor_messages_prod_batched(bx, bm, smudge_factor=1e-12):
    backend = ar.infer_backend_multi(bx, bm)
    _stack = ar.get_lib_fn(backend, "stack")

    ndim = len(bm)
    x_inds = (-1, *range(ndim))
    m_inds = [(-1, p) for p in range(ndim)]
    bmx = array_contract(
        arrays=(bx, *bm),
        inputs=(x_inds, *m_inds),
        output=x_inds,
    )

    bminv = 1 / (bm + smudge_factor)

    mouts = []
    for p in range(ndim):
        # sum all but ith index, apply inverse gate to that
        mouts.append(
            array_contract(
                arrays=(bmx, bminv[p]),
                inputs=(x_inds, m_inds[p]),
                output=m_inds[p],
            )
        )

    return _stack(mouts)


def _compute_output_single_t(
    bm,
    bx,
    normalize,
    smudge_factor=1e-12,
):
    # tensor messages
    bmo = _compute_all_tensor_messages_tree_batched(bx, bm)
    # bmo = _compute_all_tensor_messages_prod_batched(bx, bm, smudge_factor)
    normalize(bmo)
    return bmo


def _compute_output_single_m(bm, normalize, smudge_factor=1e-12):
    # index messages
    # bmo = _compute_all_hyperind_messages_tree_batched(bm)
    bmo = _compute_all_hyperind_messages_prod_batched(bm, smudge_factor)
    normalize(bmo)
    return bmo


def _update_output_to_input_single_batched(
    batched_input,
    batched_output,
    mask,
    _distance_fn,
    damping=0.0,
):
    # do a vectorized update
    select_in = (mask[0, 0, :], mask[0, 1, :], slice(None))
    select_out = (mask[1, 0, :], mask[1, 1, :], slice(None))
    bim = batched_input[select_in]
    bom = batched_output[select_out]

    # pre-damp distance
    mdiff = _distance_fn(bim, bom)

    if damping != 0.0:
        bom = damping * bim + (1 - damping) * bom

    # # post-damp distance
    # mdiff = _distance_fn(bim, bom)

    # update the input
    batched_input[select_in] = bom

    return mdiff


def _gather_zb(zb, power=1.0):
    """Given a vector of local contraction estimates `zb`, compute their
    product, avoiding underflow/overflow by accumulating the sign and exponent
    separately.

    Parameters
    ----------
    zb : array
        The local contraction estimates.
    power : float, optional
        Raise the final result to this power.

    Returns
    -------
    sign : float
        The accumulated sign or phase.
    exponent : float
        The accumulated exponent.
    """
    zb_mag = ar.do("abs", zb)
    zb_phase = zb / zb_mag

    # accumulate sign and exponent separately
    sign = ar.do("prod", zb_phase)
    exponent = ar.do("sum", ar.do("log10", zb_mag))

    if power != 1.0:
        sign **= power
        exponent *= power

    return sign, exponent


def _contract_index_region_single(bm):
    # take product over input position and sum over variable
    zb = ar.do("sum", ar.do("prod", bm, axis=0), axis=1)
    # that just leaves broadcast dimension to take product over
    return _gather_zb(zb)


def _contract_tensor_region_single(rank, batched_tensors, batched_inputs_t):
    bt = batched_tensors[rank]
    bms = batched_inputs_t[rank]
    # contract every tensor of rank `rank` with its messages
    zb = array_contract(
        [bt, *bms],
        inputs=[tuple(range(-1, rank))] + [(-1, r) for r in range(rank)],
        output=(-1,),
    )
    return _gather_zb(zb)


def _contract_messages_pair_single(
    ranki,
    ranko,
    mask,
    batched_inputs_m,
    batched_inputs_t,
):
    bmm = batched_inputs_m[ranki]
    bmt = batched_inputs_t[ranko]
    select_in = (mask[0, 0, :], mask[0, 1, :], slice(None))
    select_out = (mask[1, 0, :], mask[1, 1, :], slice(None))

    bml = bmm[select_in]
    bmr = bmt[select_out]

    zb = array_contract(
        [bml, bmr],
        inputs=[(-1, 0), (-1, 0)],
        output=(-1,),
    )

    # individual message reasons having counting factor -1
    # i.e. we are dividing by all of them
    return _gather_zb(zb, power=-1.0)


def _extract_messages_from_inputs_batched(
    batched_inputs_m,
    batched_inputs_t,
    input_locs_m,
    input_locs_t,
):
    """Get all messages as a dict from the batch stacked input form."""
    messages = {}
    for pair, (rank, p, b) in input_locs_m.items():
        messages[pair] = batched_inputs_m[rank][p, b, :]
    for pair, (rank, p, b) in input_locs_t.items():
        messages[pair] = batched_inputs_t[rank][p, b, :]
    return messages


class HV1BP(BeliefPropagationCommon):
    """Object interface for hyper, vectorized, 1-norm, belief propagation. This
    is the fast version of belief propagation possible when there are many,
    small, matching tensor sizes.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on.
    messages : dict, optional
        Initial messages to use, if not given then uniform messages are used.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    smudge_factor : float, optional
        A small number to add to the denominator of messages to avoid division
        by zero. Note when this happens the numerator will also be zero.
    thread_pool : bool or int, optional
        Whether to use a thread pool for parallelization, if ``True`` use the
        default number of threads, if an integer use that many threads.
    contract_every : int, optional
        If not None, 'contract' (via BP) the tensor network every
        ``contract_every`` iterations. The resulting values are stored in
        ``zvals`` at corresponding points ``zval_its``.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.
    """

    def __init__(
        self,
        tn,
        *,
        messages=None,
        damping=0.0,
        update="parallel",
        normalize="L2",
        distance="L2",
        smudge_factor=1e-12,
        thread_pool=False,
        contract_every=None,
        inplace=False,
    ):
        super().__init__(
            tn,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            contract_every=contract_every,
            inplace=inplace,
        )

        if update != "parallel":
            raise ValueError("Only parallel update supported.")

        self.smudge_factor = smudge_factor
        self.pool = maybe_get_thread_pool(thread_pool)
        self.initialize_messages_batched(messages)

    @property
    def normalize(self):
        return self._normalize

    @normalize.setter
    def normalize(self, normalize):
        if callable(normalize):
            # custom normalization function
            _normalize = normalize
        elif normalize == "L1":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")

            def _normalize(bx):
                bxn = _sum(_abs(bx), axis=-1, keepdims=True)
                bx /= bxn

        elif normalize == "L2":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")

            def _normalize(bx):
                bxn = _sum(_abs(bx) ** 2, axis=-1, keepdims=True) ** 0.5
                bx /= bxn

        elif normalize == "Linf":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _max = ar.get_lib_fn(self.backend, "max")

            def _normalize(bx):
                bxn = _max(_abs(bx), axis=-1, keepdims=True)
                bx /= bxn

        else:
            raise ValueError(f"Unrecognized normalize={normalize}")

        self._normalize = _normalize

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, distance):
        if callable(distance):
            # custom normalization function
            _distance_fn = distance

        elif distance == "L1":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")
            _max = ar.get_lib_fn(self.backend, "max")

            def _distance_fn(bx, by):
                return _max(_sum(_abs(bx - by), axis=-1))

        elif distance == "L2":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _sum = ar.get_lib_fn(self.backend, "sum")
            _max = ar.get_lib_fn(self.backend, "max")

            def _distance_fn(bx, by):
                return _max(_sum(_abs(bx - by) ** 2, axis=-1)) ** 0.5

        elif distance == "Linf":
            _abs = ar.get_lib_fn(self.backend, "abs")
            _max = ar.get_lib_fn(self.backend, "max")

            def _distance_fn(bx, by):
                return _max(_abs(bx - by))

        else:
            raise ValueError(f"Unrecognized distance={distance}")

        self._distance = distance
        self._distance_fn = _distance_fn

    def initialize_messages_batched(self, messages=None):
        _stack = ar.get_lib_fn(self.backend, "stack")
        _array = ar.get_lib_fn(self.backend, "array")

        if isinstance(messages, dict):
            # 'dense' (i.e. non-batch) messages explicitly supplied
            message_init_fn = None
        elif callable(messages):
            # custom message initialization function
            message_init_fn = messages
            messages = None
        elif messages == "dense":
            # explicitly create dense messages first
            message_init_fn = None
            messages = initialize_hyper_messages(self.tn)
        elif messages is None:
            # default to uniform messages
            message_init_fn = ar.get_lib_fn(self.backend, "ones")
        else:
            raise ValueError(f"Unrecognized messages={messages}")

        # here we are stacking all contractions with matching rank
        #
        #     rank: number of incident messages to a tensor or hyper index
        #     pos (p): which of those messages we are (0, 1, ..., rank - 1)
        #     batch position (b): which position in the stack we are
        #
        #     _m = messages incident to indices
        #     _t = messages incident to tensors

        # prepare index messages
        batched_inputs_m = {}
        input_locs_m = {}
        output_locs_m = {}
        shapes_m = {}

        for ix, tids in self.tn.ind_map.items():
            # all updates of the same rank can be performed simultaneously
            rank = len(tids)
            try:
                batch = batched_inputs_m[rank]
                shape = shapes_m[rank]
            except KeyError:
                batch = batched_inputs_m[rank] = [[] for _ in range(rank)]
                shape = shapes_m[rank] = [rank, 0, self.tn.ind_size(ix)]

            # batch index
            b = shape[1]
            for p, tid in enumerate(tids):
                if message_init_fn is None:
                    # we'll stack the messages later
                    batch[p].append(messages[tid, ix])
                input_locs_m[tid, ix] = (rank, p, b)
                output_locs_m[ix, tid] = (rank, p, b)

            # increment batch index
            shape[1] = b + 1

        # prepare tensor messages
        batched_tensors = {}
        batched_inputs_t = {}
        input_locs_t = {}
        output_locs_t = {}
        shapes_t = {}

        for tid, t in self.tn.tensor_map.items():
            # all updates of the same rank can be performed simultaneously
            rank = t.ndim
            if rank == 0:
                # floating scalars are not updated
                continue

            try:
                batch = batched_inputs_t[rank]
                batch_t = batched_tensors[rank]
                shape = shapes_t[rank]
            except KeyError:
                batch = batched_inputs_t[rank] = [[] for _ in range(rank)]
                batch_t = batched_tensors[rank] = []
                shape = shapes_t[rank] = [rank, 0, t.shape[0]]

            # batch index
            b = shape[1]
            for p, ix in enumerate(t.inds):
                if message_init_fn is None:
                    # we'll stack the messages later
                    batch[p].append(messages[ix, tid])
                input_locs_t[ix, tid] = (rank, p, b)
                output_locs_t[tid, ix] = (rank, p, b)

            batch_t.append(t.data)
            # increment batch index
            shape[1] = b + 1

        # combine or create batch message arrays
        for batched_inputs, shapes in zip(
            (batched_inputs_m, batched_inputs_t), (shapes_m, shapes_t)
        ):
            for rank, batch in batched_inputs.items():
                if isinstance(messages, dict):
                    # stack given messages into single arrays
                    batched_inputs[rank] = _stack(
                        tuple(_stack(batch_p) for batch_p in batch)
                    )
                else:
                    # create message arrays directly
                    batched_inputs[rank] = message_init_fn(shapes[rank])

        # stack all tensors of each rank into a single array
        for rank, tensors in batched_tensors.items():
            batched_tensors[rank] = _stack(tensors)

        # make numeric masks for updating output to input messages
        masks_m = {}
        masks_t = {}
        for masks, input_locs, output_locs in [
            (masks_m, input_locs_m, output_locs_t),
            (masks_t, input_locs_t, output_locs_m),
        ]:
            for pair in input_locs:
                (ranki, pi, bi) = input_locs[pair]
                (ranko, po, bo) = output_locs[pair]
                # we can vectorize over all distinct pairs of ranks
                key = (ranki, ranko)
                try:
                    ma_pi, ma_po, ma_bi, ma_bo = masks[key]
                except KeyError:
                    ma_pi, ma_po, ma_bi, ma_bo = masks[key] = [], [], [], []

                ma_pi.append(pi)
                ma_bi.append(bi)
                ma_po.append(po)
                ma_bo.append(bo)

            for key, (ma_pi, ma_po, ma_bi, ma_bo) in masks.items():
                # first dimension is in/out
                # second dimension is position or batch
                # third dimension is stack index
                mask = _array([[ma_pi, ma_bi], [ma_po, ma_bo]])
                masks[key] = mask

        self.batched_inputs_m = batched_inputs_m
        self.batched_inputs_t = batched_inputs_t
        self.batched_tensors = batched_tensors
        self.input_locs_m = input_locs_m
        self.input_locs_t = input_locs_t
        self.masks_m = masks_m
        self.masks_t = masks_t

    @property
    def messages(self):
        return (self.batched_inputs_m, self.batched_inputs_t)

    @messages.setter
    def messages(self, messages):
        self.batched_inputs_m, self.batched_inputs_t = messages

    def _compute_outputs_batched(
        self,
        batched_inputs,
        batched_tensors=None,
    ):
        """Given stacked messsages and optionally tensors, compute stacked
        output messages, possibly using parallel pool.
        """

        if batched_tensors is not None:
            # tensor messages
            f = _compute_output_single_t
            f_args = {
                rank: (bm, batched_tensors[rank], self.normalize)
                for rank, bm in batched_inputs.items()
            }
        else:
            # index messages
            f = _compute_output_single_m
            f_args = {
                rank: (bm, self.normalize, self.smudge_factor)
                for rank, bm in batched_inputs.items()
            }

        batched_outputs = {}
        if self.pool is None:
            # sequential process
            for rank, args in f_args.items():
                batched_outputs[rank] = f(*args)
        else:
            # parallel process
            for rank, args in f_args.items():
                batched_outputs[rank] = self.pool.submit(f, *args)
            for key, fut in batched_outputs.items():
                batched_outputs[key] = fut.result()

        return batched_outputs

    def _update_outputs_to_inputs_batched(
        self,
        batched_inputs,
        batched_outputs,
        masks,
    ):
        """Update the stacked input messages from the stacked output messages."""
        f = _update_output_to_input_single_batched
        f_args = (
            (
                batched_inputs[ranki],
                batched_outputs[ranko],
                mask,
                self._distance_fn,
                self.damping,
            )
            for (ranki, ranko), mask in masks.items()
        )

        if self.pool is None:
            # sequential process
            mdiffs = (f(*args) for args in f_args)
        else:
            # parallel process
            futs = [self.pool.submit(f, *args) for args in f_args]
            mdiffs = (fut.result() for fut in futs)

        return max(mdiffs)

    def iterate(self, tol=None):
        # first we compute new tensor output messages
        batched_outputs_t = self._compute_outputs_batched(
            batched_inputs=self.batched_inputs_t,
            batched_tensors=self.batched_tensors,
        )
        # update the index input messages with these
        t_max_mdiff = self._update_outputs_to_inputs_batched(
            self.batched_inputs_m,
            batched_outputs_t,
            self.masks_m,
        )

        # compute index messages
        batched_outputs_m = self._compute_outputs_batched(
            batched_inputs=self.batched_inputs_m,
        )
        # update the tensor input messages
        m_max_mdiff = self._update_outputs_to_inputs_batched(
            self.batched_inputs_t,
            batched_outputs_m,
            self.masks_t,
        )
        return max(t_max_mdiff, m_max_mdiff)

    def get_messages_dense(self):
        """Get messages in individual form from the batched stacks."""
        return _extract_messages_from_inputs_batched(
            self.batched_inputs_m,
            self.batched_inputs_t,
            self.input_locs_m,
            self.input_locs_t,
        )

    def get_messages(self):
        import warnings

        warnings.warn(
            "get_messages() is deprecated, or in the future it might return "
            "the batch messages, use get_messages_dense() instead.",
            DeprecationWarning,
        )

        return self.get_messages_dense()

    def contract(self, strip_exponent=False, check_zero=False):
        """Estimate the contraction of the tensor network using the current
        messages. Uses batched vectorized contractions for speed.

        Parameters
        ----------
        strip_exponent : bool, optional
            Whether to strip the exponent from the final result. If ``True``
            then the returned result is ``(mantissa, exponent)``.
        check_zero : bool, optional
            Whether to check for zero values and return zero early. Currently
            ``True`` is not implemented for HV1BP.

        Returns
        -------
        scalar or (scalar, float)
        """
        if check_zero:
            raise NotImplementedError("check_zero not implemented for HV1BP.")

        fn_args = []
        # for each rank contract index region estimate
        for bm in self.batched_inputs_m.values():
            fn_args.append((_contract_index_region_single, (bm,)))
        # for each rank contract tensor region estimate
        for rank in self.batched_tensors:
            fn_args.append(
                (
                    _contract_tensor_region_single,
                    (rank, self.batched_tensors, self.batched_inputs_t),
                )
            )
        # for each pair of ranks contract messages pair
        # region estimate which we divide by (power=-1.0)
        for (ranki, ranko), mask in self.masks_m.items():
            fn_args.append(
                (
                    _contract_messages_pair_single,
                    (
                        ranki,
                        ranko,
                        mask,
                        self.batched_inputs_m,
                        self.batched_inputs_t,
                    ),
                )
            )

        if self.pool is None:
            results = [fn(*args) for fn, args in fn_args]
        else:
            futs = [self.pool.submit(fn, *args) for fn, args in fn_args]
            results = [fut.result() for fut in futs]

        sign = 1.0
        exponent = 0.0
        for s, e in results:
            sign *= s
            exponent += e

        if strip_exponent:
            return sign, exponent
        else:
            return sign * 10**exponent

    def contract_dense(self, strip_exponent=False, check_zero=True):
        """Slow contraction via explicit extranting individual dense messages.
        This supports check_zero=True and may be useful for debugging.
        """
        return contract_hyper_messages(
            self.tn,
            self.get_messages_dense(),
            strip_exponent=strip_exponent,
            check_zero=check_zero,
            backend=self.backend,
        )


def contract_hv1bp(
    tn,
    messages=None,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    diis=False,
    update="parallel",
    normalize="L2",
    distance="L2",
    tol_abs=None,
    tol_rolling_diff=None,
    smudge_factor=1e-12,
    strip_exponent=False,
    check_zero=False,
    info=None,
    progbar=False,
):
    """Estimate the contraction of ``tn`` with hyper, vectorized, 1-norm
    belief propagation, via the exponential of the Bethe free entropy.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on, can have hyper indices.
    messages : dict, optional
        Initial messages to use, if not given then uniform messages are used.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    damping : float, optional
        The damping factor to use, 0.0 means no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    smudge_factor : float, optional
        A small number to add to the denominator of messages to avoid division
        by zero. Note when this happens the numerator will also be zero.
    strip_exponent : bool, optional
        Whether to strip the exponent from the final result. If ``True``
        then the returned result is ``(mantissa, exponent)``.
    check_zero : bool, optional
        Whether to check for zero values and return zero early.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.

    Returns
    -------
    scalar or (scalar, float)
    """
    bp = HV1BP(
        tn,
        messages=messages,
        damping=damping,
        update=update,
        normalize=normalize,
        distance=distance,
        smudge_factor=smudge_factor,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        diis=diis,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
        info=info,
        progbar=progbar,
    )
    return bp.contract(
        strip_exponent=strip_exponent,
        check_zero=check_zero,
    )


def run_belief_propagation_hv1bp(
    tn,
    messages=None,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    diis=False,
    update="parallel",
    normalize="L2",
    distance="L2",
    tol_abs=None,
    tol_rolling_diff=None,
    smudge_factor=1e-12,
    info=None,
    progbar=False,
):
    """Run belief propagation on a tensor network until it converges.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on.
    messages : dict, optional
        The current messages. For every index and tensor id pair, there should
        be a message to and from with keys ``(ix, tid)`` and ``(tid, ix)``.
        If not given, then messages are initialized as uniform.
    max_iterations : int, optional
        The maximum number of iterations to run for.
    tol : float, optional
        The convergence tolerance.
    damping : float, optional
        The damping factor to use, 0.0 means no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    smudge_factor : float, optional
        A small number to add to the denominator of messages to avoid division
        by zero. Note when this happens the numerator will also be zero.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.

    Returns
    -------
    messages : dict
        The final messages.
    converged : bool
        Whether the algorithm converged.
    """
    bp = HV1BP(
        tn,
        messages=messages,
        damping=damping,
        smudge_factor=smudge_factor,
        update=update,
        normalize=normalize,
        distance=distance,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        diis=diis,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
        info=info,
        progbar=progbar,
    )
    return bp.get_messages_dense(), bp.converged


def sample_hv1bp(
    tn,
    messages=None,
    output_inds=None,
    max_iterations=1000,
    tol=1e-2,
    damping=0.0,
    diis=False,
    update="parallel",
    normalize="L2",
    distance="L2",
    tol_abs=None,
    tol_rolling_diff=None,
    smudge_factor=1e-12,
    bias=False,
    seed=None,
    progbar=False,
):
    """Sample all indices of a tensor network using repeated belief propagation
    runs and decimation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to sample.
    messages : dict, optional
        The current messages. For every index and tensor id pair, there should
        be a message to and from with keys ``(ix, tid)`` and ``(tid, ix)``.
        If not given, then messages are initialized as uniform.
    output_inds : sequence of str, optional
        The indices to sample. If not given, then all indices are sampled.
    max_iterations : int, optional
        The maximum number of iterations for each message passing run.
    tol : float, optional
        The convergence tolerance for each message passing run.
    damping : float, optional
        The damping factor to use, 0.0 means no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    smudge_factor : float, optional
        A small number to add to each message to avoid zeros. Making this large
        is similar to adding a temperature, which can aid convergence but
        likely produces less accurate marginals.
    bias : bool or float, optional
        Whether to bias the sampling towards the largest marginal. If ``False``
        (the default), then indices are sampled proportional to their
        marginals. If ``True``, then each index is 'sampled' to be its largest
        weight value always. If a float, then the local probability
        distribution is raised to this power before sampling.
    thread_pool : bool, int or ThreadPoolExecutor, optional
        Whether to use a thread pool for parallelization. If an integer, then
        this is the number of threads to use. If ``True``, then the number of
        threads is set to the number of cores. If a ``ThreadPoolExecutor``,
        then this is used directly.
    seed : int, optional
        A random seed to use for the sampling.
    progbar : bool, optional
        Whether to show a progress bar.

    Returns
    -------
    config : dict[str, int]
        The sample configuration, mapping indices to values.
    tn_config : TensorNetwork
        The tensor network with all index values (or just those in
        `output_inds` if supllied) selected. Contracting this tensor network
        (which will just be a sequence of scalars if all index values have been
        sampled) gives the weight of the sample, e.g. should be 1 for a SAT
        problem and valid assignment.
    omega : float
        The probability of choosing this sample (i.e. product of marginal
        values). Useful possibly for importance sampling.
    """
    import numpy as np

    rng = np.random.default_rng(seed)

    tn_config = tn.copy()

    if messages is None:
        messages = initialize_hyper_messages(tn_config)

    if output_inds is None:
        output_inds = tn_config.ind_map.keys()
    output_inds = set(output_inds)

    config = {}
    omega = 1.0

    if progbar:
        import tqdm

        pbar = tqdm.tqdm(total=len(output_inds))
    else:
        pbar = None

    while output_inds:
        messages, _ = run_belief_propagation_hv1bp(
            tn_config,
            messages,
            max_iterations=max_iterations,
            tol=tol,
            damping=damping,
            diis=diis,
            update=update,
            normalize=normalize,
            distance=distance,
            tol_abs=tol_abs,
            tol_rolling_diff=tol_rolling_diff,
            smudge_factor=smudge_factor,
        )

        marginals = compute_all_index_marginals_from_messages(
            tn_config, messages
        )

        # choose most peaked marginal
        ix, p = max(
            (m for m in marginals.items() if m[0] in output_inds),
            key=lambda ix_p: max(ix_p[1]),
        )

        if bias is False:
            # sample the value according to the marginal
            v = rng.choice(np.arange(p.size), p=p)
        elif bias is True:
            v = np.argmax(p)
            # in some sense omega is really 1.0 here
        else:
            # bias towards larger marginals by raising to a power
            p = p**bias
            p /= np.sum(p)
            v = np.random.choice(np.arange(p.size), p=p)

        omega *= p[v]
        config[ix] = v

        # clean up messages
        for tid in tn_config.ind_map[ix]:
            del messages[ix, tid]
            del messages[tid, ix]

        # remove index
        tn_config.isel_({ix: v})
        output_inds.remove(ix)

        if progbar:
            pbar.update(1)
            pbar.set_description(f"{ix}->{v}", refresh=False)

    if progbar:
        pbar.close()

    return config, tn_config, omega



"""Numba accelerated tools for ranking, deranking and coupling
'flat configs', used for constructing sparse matrices in a variety of
Hilbert spaces, as well as applying operators to vectors in those spaces.


TODO:
- [ ] parallel: use threading_choose_num_blocks threading_get_block_range
      to distribute work cyclically but still in blocks
"""

import numpy as np
from numba import njit
from numba.typed import Dict

nogil = True
cache = True


@njit(cache=cache, nogil=nogil, inline="always")
def _check_next_coupled_term(
    a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
):
    """Calculate the next coupled config and coeff, or flag as zero if
    there is no coupled config. Inplace modifies the bj array.

    Parameters
    ----------
    a : uint32
        The index into the stacked terms.
    b : uint32
        The index into the stacked operators.
    n : int
        The total size of the flat configuration.
    bi : ndarray[uint8]
        The flat configuration to get the coupled configurations for.
    bj : ndarray[uint8]
        The coupled flat configuration to be modified in place.
    size_term : uint32
        The number of operators in this term.
    sizes_op : ndarray[uint8]
        Flat array of number of entries for each operator, indexed by `a`.
    regs : ndarray[uint32]
        Flat array of registers in each term, indexed by `a`.
    xis : ndarray[uint8]
        Flat array of input bits for each entry, indexed by `b`.
    xjs : ndarray[uint8]
        Flat array of output bits for each entry, indexed by `b`.
    cijs : ndarray
        Flat array of coefficients for each entry, indexed by `b`.

    Returns
    -------
    a : int
        The updated index into the stacked terms.
    b : int
        The updated index into the stacked operators.
    valid : bool
        True if the coupled config is non-zero, False otherwise.
    hij : scalar
        The coefficient for the coupled config.
    """
    # reset coeff
    hij = 1.0
    valid = True
    # reset coupled config
    for q in range(n):
        bj[q] = bi[q]

    # for each operator in the term
    for da in range(size_term):
        # the number of entries in the operator
        ia = a + da
        size_op = sizes_op[ia]

        if valid:
            # check further coupling
            reg = regs[ia]
            xi = bi[reg]

            # TODO: generalize this to more than dim=2
            if size_op == 1:
                # must match single input
                valid = xi == xis[b]
                if valid:
                    # found a match -> midstring update of coeff and config
                    bj[reg] = xjs[b]
                    hij *= cijs[b]
            else:
                # 0 and 1 both match -> midstring update of coeff and config
                ib = b + xi
                bj[reg] = xjs[ib]
                hij *= cijs[ib]

            # # alternative approaches:

            # valid = False
            # for each entry in the operator
            # for db in range(size_op):
            #     ib = b + db
            #     xin = xis[ib]
            #     if xi == xin:
            #         # found a match
            #         valid = True
            #         xj = xjs[ib]
            #         cij = cijs[ib]
            #         break

        # increment operator index
        b += size_op
    # increment the term index
    a += size_term

    return a, b, valid, hij


@njit(cache=cache, nogil=True)
def flatconfig_coupling_numba(flatconfig, coupling_map, dtype=np.float64):
    """Get the coupled flat configurations for a given flat configuration
    and coupling map.

    Like applying the sparse matrix to a basis vector, or retrieving a single
    row.

    Parameters
    ----------
    flatconfig : ndarray[uint8]
        The flat configuration to get the coupled configurations for.
    coupling_map : tuple[ndarray]
        The operator defined as tuple of flat arrays.
    dtype : {np.float64, np.complex128, np.float32, np.float64}, optional
        The dtype to use for the coupled coefficients. Default is np.float64.

    Returns
    -------
    coupled_flatconfigs : ndarray[uint8]
        A list of coupled flat configurations, each with the corresponding
        coefficient.
    coeffs : ndarray[dtype]
        The coefficients for each coupled flat configuration.
    """
    n = flatconfig.size
    buf_ptr = 0
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map
    num_terms = len(sizes_term)

    coupled_flatconfigs = np.empty((num_terms, n), dtype=np.uint8)
    coeffs = np.empty(num_terms, dtype=dtype)
    seen = Dict.empty(np.int64, np.int64)

    # indices into stacked terms and operators
    a = b = np.uint32(0)
    for size_term in sizes_term:
        # for each term in the hamiltonian find which, if
        # any, config it couples to, & with what coefficient
        a, b, valid, hij = _check_next_coupled_term(
            a, b, n, flatconfig, bj, size_term, sizes_op, regs, xis, xjs, cijs
        )
        if valid:
            key = flatconfig_to_rank_nosymm(bj)
            if key in seen:
                # already seen this config -> simply add to coeff
                old_ptr = seen[key]
                coeffs[old_ptr] += hij
            else:
                seen[key] = buf_ptr
                coupled_flatconfigs[buf_ptr, :] = bj
                coeffs[buf_ptr] = hij
                buf_ptr += 1

    return coupled_flatconfigs[:buf_ptr], coeffs[:buf_ptr]


# ----------------------- unconstrained hilbert space ----------------------- #


@njit(cache=cache, nogil=nogil, inline="always")
def flatconfig_to_rank_nosymm(flatconfig):
    """Convert a flat config array to a rank, i.e. its position in the
    lexicographic ordering of all bitstrings of that length.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8.

    Returns
    -------
    int
    """
    r = 0
    for xi in flatconfig:
        r = (r << 1) | xi
    return r


@njit(cache=cache, nogil=nogil, inline="always")
def rank_into_flatconfig_nosymm(flatconfig, r, n):
    """Inplace conversion of a rank to a flat config array.

    Parameters
    ----------
    flatconfig : array_like
        A input flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8, it will be overwritten.
    r : int
        The rank to convert.
    n : int
        The total length of the flat config array.
    """
    for i in range(n - 1, -1, -1):
        flatconfig[i] = r & 1
        r >>= 1


@njit(cache=cache, nogil=nogil, inline="always")
def rank_to_flatconfig_nosymm(r, n):
    """Convert a rank to a flat config array.

    Parameters
    ----------
    r : int
        The rank to convert.
    n : int
        The total length of the flat config array.

    Returns
    -------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array will be of dtype uint8.
    """
    flatconfig = np.empty(n, dtype=np.uint8)
    rank_into_flatconfig_nosymm(flatconfig, r, n)
    return flatconfig


@njit(cache=cache, nogil=nogil)
def build_coo_numba_core_nosymm(
    n,
    coupling_map,
    dtype=np.float64,
    world_size=1,
    world_rank=0,
):
    """Build sparse coo data in a unconstrained hilbert space."""
    D = 2**n

    buf_size = D
    data = np.empty(buf_size, dtype=dtype)
    rows = np.empty(buf_size, dtype=np.int64)
    cols = np.empty(buf_size, dtype=np.int64)
    buf_ptr = 0

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_nosymm(bi, ci, n)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                if buf_ptr >= buf_size:
                    # need to double our storage
                    data = np.concatenate((data, np.empty_like(data)))
                    rows = np.concatenate((rows, np.empty_like(rows)))
                    cols = np.concatenate((cols, np.empty_like(cols)))
                    buf_size *= 2

                # convert coupled config back to rank
                cj = flatconfig_to_rank_nosymm(bj)

                data[buf_ptr] = hij
                cols[buf_ptr] = ci
                rows[buf_ptr] = cj
                buf_ptr += 1

    return data[:buf_ptr], rows[:buf_ptr], cols[:buf_ptr]


@njit(cache=cache, nogil=nogil)
def matvec_nosymm(
    x,
    out,
    n,
    coupling_map,
    world_size=1,
    world_rank=0,
):
    D = 2**n

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_nosymm(bi, ci, n)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                cj = flatconfig_to_rank_nosymm(bj)
                out[cj] += hij * x[ci]


# ----------------- mixed-radix unconstrained hilbert space ----------------- #


@njit(cache=cache, nogil=nogil, inline="always")
def calculate_strides(sizes):
    n = len(sizes)
    strides = np.ones(n, dtype=np.uint64)
    for i in range(n - 2, -1, -1):
        strides[i] = strides[i + 1] * sizes[i + 1]
    return strides


@njit(cache=cache, nogil=nogil, inline="always")
def flatconfig_to_rank_mixed_radix_nosymm(flatconfig, strides):
    """Convert a flat config array for a mixed size system to a rank, i.e. its
    position in the lexicographic ordering of all configurations.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a configuration.
        The array should be of dtype uint8.
    strides : array_like
        The strides array of shape (n,) The array should be of dtype uint64.

    Returns
    -------
    np.uint64
    """
    r = 0
    for i in range(flatconfig.size):
        r += flatconfig[i] * strides[i]
    return r


@njit(cache=cache, nogil=nogil, inline="always")
def rank_into_flatconfig_mixed_radix_nosymm(flatconfig, r, sizes, strides):
    """Inplace conversion of a mixed-radix rank to a flat config array.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a configuration.
        The array should be of dtype uint8.
    r : np.uint64
        The rank to convert.
    sizes : array_like
        The sizes of the configuration dimensions.
    strides : array_like
        The strides of the configuration dimensions.
    """
    for i in range(len(sizes)):
        flatconfig[i] = (r // strides[i]) % sizes[i]


@njit(cache=cache, nogil=nogil, inline="always")
def rank_to_flatconfig_mixed_radix_nosymm(
    r,
    sizes,
    strides,
) -> np.ndarray:
    """Convert a mixed-radix rank to a flat config array.

    Parameters
    ----------
    r : np.uint64
        The rank to convert.
    sizes : array_like
        The sizes of the configuration dimensions.
    strides : array_like
        The strides of the configuration dimensions.

    Returns
    -------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a configuration.
        The array will be of dtype uint8.
    """
    n = len(sizes)
    flatconfig = np.empty(n, dtype=np.uint8)
    rank_into_flatconfig_mixed_radix_nosymm(flatconfig, r, sizes, strides)
    return flatconfig


# --------------------- parity conserved hilbert space ---------------------- #


@njit(cache=cache, nogil=nogil, inline="always")
def flatconfig_to_rank_z2(flatconfig):
    """Convert a flat config array to a z2 rank, i.e. its position in the
    lexicographic ordering of all bitstrings of even or odd parity.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8.

    Returns
    -------
    int
    """
    r = 0
    # we treat like the nosymm case and simply ignore the last bit
    for i in range(flatconfig.size - 1):
        r = (r << 1) | flatconfig[i]
    return r


@njit(cache=cache, nogil=nogil, inline="always")
def rank_into_flatconfig_z2(flatconfig, r, n, p):
    """Inplace conversion of a z2 rank to a flat config array.

    Parameters
    ----------
    flatconfig : array_like
        A input flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8, it will be overwritten.
    r : int
        The rank to convert.
    n : int
        The total length of the flat config array.
    """
    prem = 0
    m = 1 << (n - 2)
    for i in range(n - 1):
        xi = r & m != 0
        flatconfig[i] = xi
        m >>= 1
        prem ^= xi
    flatconfig[n - 1] = prem ^ p


@njit(cache=cache, nogil=nogil, inline="always")
def rank_to_flatconfig_z2(r, n, p):
    """Convert a z2 rank to a flat config array.

    Parameters
    ----------
    r : int
        The rank to convert.
    n : int
        The total length of the flat config array.
    p : int
        The parity of the flat config array.

    Returns
    -------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array will be of dtype uint8.
    """
    flatconfig = np.empty(n, dtype=np.uint8)
    rank_into_flatconfig_z2(flatconfig, r, n, p)
    return flatconfig


@njit(cache=cache, nogil=nogil)
def build_coo_numba_core_z2(
    n,
    p,
    coupling_map,
    dtype=np.float64,
    world_size=1,
    world_rank=0,
):
    """Build sparse coo data in a parity conserved hilbert space."""
    D = 2 ** (n - 1)

    buf_size = D
    data = np.empty(buf_size, dtype=dtype)
    rows = np.empty(buf_size, dtype=np.int64)
    cols = np.empty(buf_size, dtype=np.int64)
    buf_ptr = 0

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_z2(bi, ci, n, p)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                if buf_ptr >= buf_size:
                    # need to double our storage
                    data = np.concatenate((data, np.empty_like(data)))
                    rows = np.concatenate((rows, np.empty_like(rows)))
                    cols = np.concatenate((cols, np.empty_like(cols)))
                    buf_size *= 2

                # convert coupled config back to rank
                cj = flatconfig_to_rank_z2(bj)

                data[buf_ptr] = hij
                cols[buf_ptr] = ci
                rows[buf_ptr] = cj
                buf_ptr += 1

    return data[:buf_ptr], rows[:buf_ptr], cols[:buf_ptr]


@njit(cache=cache, nogil=nogil)
def matvec_z2(
    x,
    out,
    n,
    p,
    coupling_map,
    world_size=1,
    world_rank=0,
):
    D = 2 ** (n - 1)

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_z2(bi, ci, n, p)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                cj = flatconfig_to_rank_z2(bj)
                out[cj] += hij * x[ci]


# -------------------- particle conserved hilbert space --------------------- #


@njit(cache=cache, nogil=nogil)
def build_pascal_table(nmax):
    """Build the Pascal triangle table for the number of ways to choose k
    elements from n elements, i.e. the binomial coefficients.
    The table is of shape (nmax+1, nmax+1) and the entry at (n, k) is
    the number of ways to choose k elements from n elements.

    Parameters
    ----------
    nmax : int
        The maximum number of particles.

    Returns
    -------
    pt : array_like
        The Pascal triangle table of shape (nmax+1, nmax+1).
    """
    d = nmax + 1
    pt = np.zeros((d, d), dtype=np.int64)
    for n in range(d):
        pt[n, 0] = 1
        for k in range(1, n + 1):
            pt[n, k] = pt[n - 1, k - 1] + pt[n - 1, k]
    return pt


@njit(cache=cache, nogil=nogil, inline="always")
def flatconfig_to_rank_u1_pascal(flatconfig, n, k, pt):
    """Given a flat config array, return the u1 rank of the config in the
    lexicographic ordering of all bitstrings of that length with hamming weight
    k.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8.
    n : int
        The total length of the flat config array.
    k : int
        The number of particles in the flat config array.
    pt : array_like
        The Pascal triangle table of shape containing at least (n, k).

    Returns
    -------
    r : int
        The rank of the config in the lexicographic ordering.
    """
    r = 0
    krem = k
    j = n
    for i, xi in enumerate(flatconfig):
        j -= 1
        r += xi * pt[j, krem]
        krem -= xi
    return r


@njit(cache=cache, nogil=nogil, inline="always")
def rank_into_flatconfig_u1_pascal(flatconfig, r, n, k, pt):
    """Inplace conversion of a rank to a flat config array with hamming weight
    k.

    Parameters
    ----------
    flatconfig : array_like
        A input flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8, it will be overwritten.
    r : int
        The rank to convert.
    n : int
        The total length of the flat config array.
    k : int
        The number of particles in the flat config array.
    pt : array_like
        The Pascal triangle table of shape containing at least (n, k).
    """
    r = np.int64(r)
    krem = np.int64(k)
    j = n
    for i in range(n):
        j -= 1
        rank_if_one = pt[j, krem]
        if r >= rank_if_one:
            flatconfig[i] = 1
            r -= rank_if_one
            krem -= 1
        else:
            flatconfig[i] = 0


@njit(cache=cache, nogil=nogil, inline="always")
def rank_to_flatconfig_u1_pascal(r, n, k, pt):
    """Convert a rank to a flat config array with hamming weight k.

    Parameters
    ----------
    r : int
        The rank to convert.
    n : int
        The total length of the flat config array.
    k : int
        The number of particles in the flat config array.
    pt : array_like
        The Pascal triangle table of shape containing at least (n, k).

    Returns
    -------
    array_like
        A flat config array of shape (n,), corresponding to a bitstring.
    """
    flatconfig = np.empty(n, dtype=np.uint8)
    rank_into_flatconfig_u1_pascal(flatconfig, r, n, k, pt)
    return flatconfig


@njit(cache=cache, nogil=nogil)
def build_coo_numba_core_u1(
    n,
    k,
    coupling_map,
    dtype=np.float64,
    world_size=1,
    world_rank=0,
):
    """Build sparse coo data in a number conserved hilbert space."""
    pt = build_pascal_table(n)
    D = pt[n, k]

    buf_size = D
    data = np.empty(buf_size, dtype=dtype)
    rows = np.empty(buf_size, dtype=np.int64)
    cols = np.empty(buf_size, dtype=np.int64)
    buf_ptr = 0

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_u1_pascal(bi, ci, n, k, pt)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                if buf_ptr >= buf_size:
                    # need to double our storage
                    data = np.concatenate((data, np.empty_like(data)))
                    rows = np.concatenate((rows, np.empty_like(rows)))
                    cols = np.concatenate((cols, np.empty_like(cols)))
                    buf_size *= 2

                # convert coupled config back to rank
                cj = flatconfig_to_rank_u1_pascal(bj, n, k, pt)

                data[buf_ptr] = hij
                cols[buf_ptr] = ci
                rows[buf_ptr] = cj
                buf_ptr += 1

    return data[:buf_ptr], rows[:buf_ptr], cols[:buf_ptr]


@njit(cache=cache, nogil=nogil)
def matvec_u1(
    x,
    out,
    n,
    k,
    coupling_map,
    world_size=1,
    world_rank=0,
):
    pt = build_pascal_table(n)
    D = pt[n, k]

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_u1_pascal(bi, ci, n, k, pt)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                cj = flatconfig_to_rank_u1_pascal(bj, n, k, pt)
                out[cj] += hij * x[ci]


# --------------------- doubly conserved hilbert space ---------------------- #


@njit(cache=cache, nogil=nogil, inline="always")
def flatconfig_to_rank_u1u1_pascal(flatconfig, na, ka, nb, kb, pt):
    """Convert a flat config array to a doubly number conserved rank, i.e.
    its position in the ordering of all bitstrings of lenght `na + nb` with
    hamming weight `ka` and `kb` on each section respectively.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (na + nb,), corresponding to a bitstring.
        The array should be of dtype uint8.
    na : int
        The total length of the first section of the flat config array.
    ka : int
        The hamming weight of the first section.
    nb : int
        The total length of the second section of the flat config array.
    kb : int
        The hamming weight of the second section.
    pt : array_like
        The Pascal triangle table of shape containing at least max(na, nb).

    Returns
    -------
    r : int
        The rank of the config in the lexicographic ordering.
    """
    Db = pt[nb, kb]
    return flatconfig_to_rank_u1_pascal(
        flatconfig[:na], na, ka, pt
    ) * Db + flatconfig_to_rank_u1_pascal(flatconfig[na:], nb, kb, pt)


@njit(cache=cache, nogil=nogil, inline="always")
def rank_into_flatconfig_u1u1_pascal(flatconfig, r, na, ka, nb, kb, pt):
    """Inplace conversion of a doubly number conserved rank to a flat config
    array.

    Parameters
    ----------
    flatconfig : array_like
        A input flat config array of shape (na + nb,), corresponding to a
        bitstring. The array should be of dtype uint8, it will be overwritten.
    r : int
        The rank to convert.
    na : int
        The total length of the first section of the flat config array.
    ka : int
        The hamming weight of the first section.
    nb : int
        The total length of the second section of the flat config array.
    kb : int
        The hamming weight of the second section.
    pt : array_like
        The Pascal triangle table of shape containing at least max(na, nb).
    """
    Db = pt[nb, kb]
    r1 = r // Db
    r2 = r % Db
    rank_into_flatconfig_u1_pascal(flatconfig[:na], r1, na, ka, pt)
    rank_into_flatconfig_u1_pascal(flatconfig[na:], r2, nb, kb, pt)


@njit(cache=cache, nogil=nogil, inline="always")
def rank_to_flatconfig_u1u1_pascal(r, na, ka, nb, kb, pt):
    """Convert a doubly number conserved rank to a flat config array.

    Parameters
    ----------
    r : int
        The rank to convert.
    na : int
        The total length of the first section of the flat config array.
    ka : int
        The hamming weight of the first section.
    nb : int
        The total length of the second section of the flat config array.
    kb : int
        The hamming weight of the second section.
    pt : array_like
        The Pascal triangle table of shape containing at least max(na, nb).

    Returns
    -------
    flatconfig : array_like
        A flat config array of shape (na + nb,), corresponding to a bitstring.
        The array will be of dtype uint8.
    """
    flatconfig = np.empty(na + nb, dtype=np.uint8)
    rank_into_flatconfig_u1u1_pascal(flatconfig, r, na, ka, nb, kb, pt)
    return flatconfig


@njit(cache=cache, nogil=nogil)
def build_coo_numba_core_u1u1(
    na,
    ka,
    nb,
    kb,
    coupling_map,
    dtype=np.float64,
    world_size=1,
    world_rank=0,
):
    pt = build_pascal_table(max(na, nb))
    D = pt[na, ka] * pt[nb, kb]
    n = na + nb

    buf_size = D
    data = np.empty(buf_size, dtype=dtype)
    rows = np.empty(buf_size, dtype=np.int64)
    cols = np.empty(buf_size, dtype=np.int64)
    buf_ptr = 0

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting configs
        rank_into_flatconfig_u1u1_pascal(bi, ci, na, ka, nb, kb, pt)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                if buf_ptr >= buf_size:
                    # need to double our storage
                    data = np.concatenate((data, np.empty_like(data)))
                    rows = np.concatenate((rows, np.empty_like(rows)))
                    cols = np.concatenate((cols, np.empty_like(cols)))
                    buf_size *= 2

                # convert coupled config back to rank
                cj = flatconfig_to_rank_u1u1_pascal(bj, na, ka, nb, kb, pt)

                data[buf_ptr] = hij
                cols[buf_ptr] = ci
                rows[buf_ptr] = cj
                buf_ptr += 1

    return data[:buf_ptr], rows[:buf_ptr], cols[:buf_ptr]


@njit(cache=cache, nogil=nogil)
def matvec_u1u1(
    x,
    out,
    na,
    ka,
    nb,
    kb,
    coupling_map,
    world_size=1,
    world_rank=0,
):
    pt = build_pascal_table(max(na, nb))
    D = pt[na, ka] * pt[nb, kb]
    n = na + nb

    bi = np.empty(n, dtype=np.uint8)
    bj = np.empty(n, dtype=np.uint8)

    sizes_term, regs, sizes_op, xis, xjs, cijs = coupling_map

    for ci in range(world_rank, D, world_size):
        # reset the starting config
        rank_into_flatconfig_u1u1_pascal(bi, ci, na, ka, nb, kb, pt)
        # indices into stacked terms and operators
        a = b = np.uint32(0)
        for size_term in sizes_term:
            # for each term in the hamiltonian find which, if
            # any, config it couples to, & with what coefficient
            a, b, valid, hij = _check_next_coupled_term(
                a, b, n, bi, bj, size_term, sizes_op, regs, xis, xjs, cijs
            )
            if valid:
                cj = flatconfig_to_rank_u1u1_pascal(bj, na, ka, nb, kb, pt)
                out[cj] += hij * x[ci]


# ------------------------------- public api -------------------------------- #


@njit(cache=cache, nogil=nogil)
def rank_to_flatconfig(r, sector, symmetry=0, pt=None):
    """Convert a rank to a flat config array.

    Parameters
    ----------
    r : int
        The rank to convert.
    sector : tuple[int]
        Specifies the sector to convert.

        - (n,) for unconstrained hilbert space
        - (n, parity) for Z2 symmetry
        - (n, k) for U1 symmetry
        - (na, ka, nb, kb) for U1U1 symmetry

    symmetry : {0, 1, 2, 3}, optional
        Specifies the symmetry to use. 0 = "None", 1 = "Z2", 2 = "U1",
        3 = "U1U1". Default is 0.
    pt : array_like, optional
        The Pascal triangle table of shape containing at least max(na, nb).
        If not provided, it will be built internally. This is only used for
        U1 and U1U1 hilbert spaces.

    Returns
    -------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array will be of dtype uint8.
    """
    if symmetry == 0:
        # unconstrained hilbert space
        (n,) = sector
        return rank_to_flatconfig_nosymm(r, n)
    elif symmetry == 1:
        n, p = sector
        if pt is None:
            pt = build_pascal_table(n)
        return rank_to_flatconfig_z2(r, n, p)
    elif symmetry == 2:
        n, k = sector
        if pt is None:
            pt = build_pascal_table(n)
        return rank_to_flatconfig_u1_pascal(r, n, k, pt)
    elif symmetry == 3:
        na, ka, nb, kb = sector
        if pt is None:
            pt = build_pascal_table(max(na, nb))
        return rank_to_flatconfig_u1u1_pascal(r, na, ka, nb, kb, pt)
    else:
        raise ValueError(
            f"Symmetry must be 0, 1, 2, or 3. Got {symmetry} instead."
        )


@njit(cache=cache, nogil=nogil)
def flatconfig_to_rank(flatconfig, sector, symmetry=0, pt=None):
    """Convert a flat config array to a rank, i.e. its position in the
    lexicographic ordering of all bitstrings of that length.

    Parameters
    ----------
    flatconfig : array_like
        A flat config array of shape (n,), corresponding to a bitstring.
        The array should be of dtype uint8.
    sector : tuple[int]
        Specifies the sector to convert.

        - (n,) for unconstrained hilbert space
        - (n, parity) for Z2 symmetry
        - (n, k) for U1 symmetry
        - (na, ka, nb, kb) for U1U1 symmetry

    symmetry : {0, 1, 2, 3}, optional
        Specifies the symmetry to use. 0 = "None", 1 = "Z2", 2 = "U1",
        3 = "U1U1". Default is 0.
    pt : array_like, optional
        The Pascal triangle table of shape containing at least max(na, nb).
        If not provided, it will be built internally. This is only used for
        U1 and U1U1 hilbert spaces.

    Returns
    -------
    int
    """
    if symmetry == 0:
        # unconstrained hilbert space
        (n,) = sector
        return flatconfig_to_rank_nosymm(flatconfig)
    elif symmetry == 1:
        n, p = sector
        return flatconfig_to_rank_z2(flatconfig, n, p)
    elif symmetry == 2:
        n, k = sector
        if pt is None:
            pt = build_pascal_table(n)
        return flatconfig_to_rank_u1_pascal(flatconfig, n, k, pt)
    elif symmetry == 3:
        na, ka, nb, kb = sector
        if pt is None:
            pt = build_pascal_table(max(na, nb))
        return flatconfig_to_rank_u1u1_pascal(flatconfig, na, ka, nb, kb, pt)
    else:
        raise ValueError(
            f"Symmetry must be 0, 1, 2, or 3. Got {symmetry} instead."
        )


@njit(cache=cache, nogil=nogil)
def build_coo_numba_core(
    coupling_map,
    sector,
    symmetry=0,
    dtype=np.float64,
    world_size=1,
    world_rank=0,
):
    """Build the data for a sparse matrix in COO format.

    Parameters
    ----------
    coupling_map : tuple[ndarray]
        The operator defined as tuple of flat arrays.
    sector : tuple[int]
        Specifies the sector to convert.

        - (n,) for unconstrained hilbert space
        - (n, parity) for Z2 symmetry
        - (n, k) for U1 symmetry
        - (na, ka, nb, kb) for U1U1 symmetry

    symmetry : {0, 1, 2, 3}, optional
        Specifies the symmetry to use. 0 = "None", 1 = "Z2", 2 = "U1",
        3 = "U1U1". Default is 0.
    dtype : {np.float64, np.complex128, np.float32, np.float64}, optional
        The dtype to use for the data. Default is np.float64.
    world_size : int, optional
        The number of processes in the world. Default is 1. Only rows
        corresponding to range(world_rank, D, world_size) will be computed.
        This is used for parallelization.
    world_rank : int, optional
        The rank of the current process. Default is 0. Only rows
        corresponding to range(world_rank, D, world_size) will be computed.
        This is used for parallelization.

    Returns
    -------
    data : ndarray[float64]
        The data for the sparse matrix in COO format.
    rows : ndarray[int64]
        The row indices for the sparse matrix in COO format.
    cols : ndarray[int64]
        The column indices for the sparse matrix in COO format.
    """
    if symmetry == 0:
        # unconstrained qubit hilbert space
        (n,) = sector
        return build_coo_numba_core_nosymm(
            n, coupling_map, dtype, world_size, world_rank
        )
    elif symmetry == 1:
        n, p = sector
        return build_coo_numba_core_z2(
            n, p, coupling_map, dtype, world_size, world_rank
        )

    elif symmetry == 2:
        n, k = sector
        return build_coo_numba_core_u1(
            n, k, coupling_map, dtype, world_size, world_rank
        )
    elif symmetry == 3:
        na, ka, nb, kb = sector
        return build_coo_numba_core_u1u1(
            na, ka, nb, kb, coupling_map, dtype, world_size, world_rank
        )
    else:
        raise ValueError(
            f"Symmetry must be 0, 1, 2, or 3. Got {symmetry} instead."
        )


@njit(cache=cache, nogil=nogil)
def matvec_numba(
    x,
    out,
    coupling_map,
    sector,
    symmetry=0,
    world_size=1,
    world_rank=0,
):
    """Apply the operator defined by the coupling map to the input vector.

    Parameters
    ----------
    x : ndarray[float64]
        The input vector to apply the operator to.
    out : ndarray[float64]
        The output vector to store the result.
    coupling_map : tuple[ndarray]
        The operator defined as tuple of flat arrays.
    sector : tuple[int]
        Specifies the sector to convert.

        - (n,) for unconstrained hilbert space
        - (n, parity) for Z2 symmetry
        - (n, k) for U1 symmetry
        - (na, ka, nb, kb) for U1U1 symmetry

    symmetry : {0, 1, 2, 3}, optional
        Specifies the symmetry to use. 0 = "None", 1 = "Z2", 2 = "U1",
        3 = "U1U1". Default is 0.
    world_size : int, optional
        The number of processes in the world. Default is 1. Only rows
        corresponding to range(world_rank, D, world_size) will be computed.
        This is used for parallelization.
    world_rank : int, optional
        The rank of the current process. Default is 0. Only rows
        corresponding to range(world_rank, D, world_size) will be computed.
        This is used for parallelization.
    """
    if symmetry == 0:
        # unconstrained hilbert space
        (n,) = sector
        matvec_nosymm(x, out, n, coupling_map, world_size, world_rank)
    elif symmetry == 1:
        n, p = sector
        matvec_z2(x, out, n, p, coupling_map, world_size, world_rank)
    elif symmetry == 2:
        n, k = sector
        matvec_u1(x, out, n, k, coupling_map, world_size, world_rank)
    elif symmetry == 3:
        na, ka, nb, kb = sector
        matvec_u1u1(
            x,
            out,
            na,
            ka,
            nb,
            kb,
            coupling_map,
            world_size,
            world_rank,
        )
    else:
        raise ValueError(
            f"Symmetry must be 0, 1, 2, or 3. Got {symmetry} instead."
        )



import array

from autoray import do
from quimb.tensor import TensorNetwork
from quimb.tensor.decomp import compute_bondenv_projectors
from quimb.tensor.tensor_arbgeom import TensorNetworkGenVector
from quimb.tensor.tensor_arbgeom_tebd import SimpleUpdateGen
from quimb.tensor.tensor_core import tensor_make_single_bond


def compress_between_tids_bondenv_exact(
    self: TensorNetwork,
    tida,
    tidb,
    max_bond,
    cutoff=0.0,
    absorb="both",
    max_iterations=100,
    tol=1e-10,
    solver="solve",
    solver_maxiter=4,
    prenormalize=False,
    condition=True,
    enforce_pos=True,
    pos_smudge=1e-10,
    init="svd",
    info=None,
    optimize="auto-hq",
    **contract_opts,
):
    """Compress the bond between the two tensors identified by ``tida`` and
    ``tidb`` exactly, by computing the full bond environment tensor and
    iteratively fitting compressed (low-rank) projectors to it.

    Parameters
    ----------
    tida : int
        The identifier of the first tensor.
    tidb : int
        The identifier of the second tensor.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        The singular value cutoff to use.
    absorb : {'both', 'left', 'right', None}, optional
        How to absorb the effective singular values into the tensors.
    max_iterations : int, optional
        The maximum number of iterations to use when fitting the projectors.
    tol : float, optional
        The target tolerance to reach when fitting the projectors.
    solver : {'solve', None, str}, optional
        The solver to use inside the fitting loop. If None will use a custom
        conjugate gradient method. Else can be any of the iterative solvers
        supported by ``scipy.sparse.linalg`` such as 'gmres', 'bicgstab', etc.
    solver_maxiter : int, optional
        The maximum number of iterations to use for the *inner* solver, i.e.
        per fitting step, only for iterative `solver` args.
    prenormalize : bool, optional
        Whether to prenormalize the environment tensor such that its full
        contraction before compression is 1. Recommended for stability when
        the normalization does not matter.
    condition : bool or "iso", optional
        Whether to condition the projectors after each fitting step. If
        ``True``, their norms will be simply matched. If ``"iso"``, then they
        are gauged each time such that the previous tensor is isometric.
        Recommended for stability.
    enforce_pos : bool, optional
        Whether to enforce the environment tensor to be positive semi-definite
        by symmetrizing and clipping negative eigenvalues. Recommended for
        stability.
    pos_smudge : float, optional
        The value to clip negative eigenvalues to when enforcing positivity,
        relative to the largest eigenvalue.
    init : {'svd', 'eigh', 'random', 'reduced'}, optional
        How to initialize the compression projectors. The options are:

        - 'svd': use a truncated SVD of the environment tensor with the bra
          bond traced out.
        - 'eigh': use a similarity compression of the environment tensor with
          the bra bond traced out.
        - 'random': use random projectors.
        - 'reduced': split the environment into bra and ket parts, then
          canonize one half left and right to get the reduced factors.

    info : dict, optional
        If provided, will store information about the fitting process here.
        The keys 'iterations' and 'distance' will contain the final number of
        iterations and distance reached respectively.
    optimize : str, optional
        Contraction path optimizer to use when forming the bond environment.
    contract_opts
        Other contraction options to pass.
    """
    ta = self.tensor_map[tida]
    tb = self.tensor_map[tidb]

    # get and cut bond
    _, bix, _ = tensor_make_single_bond(ta, tb)
    ta.reindex_({bix: "kl"})
    tb.reindex_({bix: "kr"})

    # contract the 4-index bond tensor
    b = self.conj().reindex_({"kl": "bl", "kr": "br"})
    E = (
        (self | b)
        .contract(
            all,
            output_inds=("kl", "kr", "bl", "br"),
            optimize=optimize,
            **contract_opts,
        )
        .data
    )

    # compute the projectors
    Pl, svals, Pr = compute_bondenv_projectors(
        E,
        max_bond=max_bond,
        cutoff=cutoff,
        max_iterations=max_iterations,
        tol=tol,
        absorb=absorb,
        solver=solver,
        solver_maxiter=solver_maxiter,
        prenormalize=prenormalize,
        condition=condition,
        enforce_pos=enforce_pos,
        pos_smudge=pos_smudge,
        init=init,
        info=info,
    )

    # absorb the projectors to compress the tn!
    ta.gate_(Pl, "kl", transposed=True)
    tb.gate_(Pr, "kr")

    # reconnect bond
    ta.reindex_({"kl": bix})
    tb.reindex_({"kr": bix})

    if (svals is not None) and (info is not None):
        # store the singular values
        info["singular_values"] = svals / do("linalg.norm", svals)


def compress_between_tids_bondenv_cluster(
    self: TensorNetwork,
    tida,
    tidb,
    max_bond,
    cutoff=0.0,
    gauges=None,
    max_distance=1,
    mode="graphdistance",
    fillin=False,
    grow_from="all",
    max_iterations=100,
    tol=1e-10,
    solver="solve",
    solver_maxiter=4,
    prenormalize=False,
    condition=True,
    enforce_pos=True,
    pos_smudge=1e-10,
    init="svd",
    gauge_power=1.0,
    gauge_smudge=1e-10,
    optimize="auto-hq",
    info=None,
    **contract_opts,
):
    """Compress the bond between the two tensors identified by ``tida`` and
    ``tidb`` using a cluster of tensors around them to approximate the bond
    environment tensor.

    Parameters
    ----------
    tida : int
        The identifier of the first tensor.
    tidb : int
        The identifier of the second tensor.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        The singular value cutoff to use, once the compressed projectors
        have been fitted.
    gauges : dict[any, array], optional
        Gauges, in terms of singular value vectors, to absorb around the
        cluster.
    max_distance : int, optional
        The maximum distance to the initial tagged region, or if using
        'loopunion' mode, the maximum size of any loop.
    mode : {'graphdistance', 'loopunion'}, optional
        How to select the local tensors, either by graph distance or by
        selecting the union of all loopy regions containing ``tids``.
    fillin : bool or int, optional
        Whether to fill in the local patch with additional tensors, or not.
        `fillin` tensors are those connected by two or more bonds to the
        original local patch, the process is repeated int(fillin) times.
    grow_from : {"all", "any", "alldangle", "anydangle"}, optional
        If mode is 'loopunion', whether each loop should contain *all* of
        the initial tids, or just *any* of them (generating a larger
        region). If 'alldangle' or 'anydangle', then the individual
        loops can contain the target tids even if they are dangling.
    max_iterations : int, optional
        The maximum number of iterations to use when fitting the projectors.
    tol : float, optional
        The target tolerance to reach when fitting the projectors.
    solver : {'solve', None, str}, optional
        The solver to use inside the fitting loop. If None will use a custom
        conjugate gradient method. Else can be any of the iterative solvers
        supported by ``scipy.sparse.linalg`` such as 'gmres', 'bicgstab', etc.
    solver_maxiter : int, optional
        The maximum number of iterations to use for the *inner* solver, i.e.
        per fitting step, only for iterative `solver` args.
    prenormalize : bool, optional
        Whether to prenormalize the environment tensor such that its full
        contraction before compression is 1. Recommended for stability when
        the normalization does not matter.
    condition : bool or "iso", optional
        Whether to condition the projectors after each fitting step. If
        ``True``, their norms will be simply matched. If ``"iso"``, then they
        are gauged each time such that the previous tensor is isometric.
        Recommended for stability.
    enforce_pos : bool, optional
        Whether to enforce the environment tensor to be positive semi-definite
        by symmetrizing and clipping negative eigenvalues. Recommended for
        stability.
    pos_smudge : float, optional
        The value to clip negative eigenvalues to when enforcing positivity,
        relative to the largest eigenvalue.
    init : {'svd', 'eigh', 'random', 'reduced'}, optional
        How to initialize the compression projectors. The options are:

        - 'svd': use a truncated SVD of the environment tensor with the bra
          bond traced out.
        - 'eigh': use a similarity compression of the environment tensor with
          the bra bond traced out.
        - 'random': use random projectors.
        - 'reduced': split the environment into bra and ket parts, then
          canonize one half left and right to get the reduced factors.

    gauge_power : float, optional
        A power to raise the gauge vectors to when inserting.
    gauge_smudge : float, optional
        A small value to add to the gauge vectors to avoid singularities
        when inserting.
    optimize : str, optional
        Contraction path optimizer to use when forming the bond environment.
    info : dict, optional
        If provided, will store information about the fitting process here.
        The keys 'iterations' and 'distance' will contain the final number of
        iterations and distance reached respectively.
    contract_opts
        Other contraction options to pass.
    """
    if info is None:
        info = {}

    if gauges is not None:
        # handle the target bond explicitly, as it is changing size
        ta = self.tensor_map[tida]
        tb = self.tensor_map[tidb]
        _, bix, _ = tensor_make_single_bond(ta, tb)
        g = gauges.pop(bix, None)
        if g is not None:
            gsqrt = do("sqrt", g)
            ta.multiply_index_diagonal_(bix, gsqrt)
            tb.multiply_index_diagonal_(bix, gsqrt)

        # we will extract new gauge directly
        absorb = None
    else:
        absorb = "both"

    k = self._select_local_tids(
        (tida, tidb),
        max_distance=max_distance,
        mode=mode,
        fillin=fillin,
        grow_from=grow_from,
    )

    if gauges is not None:
        outer, inner = k.gauge_simple_insert(
            gauges,
            power=gauge_power,
            smudge=gauge_smudge,
        )
    else:
        outer = inner = None

    compress_between_tids_bondenv_exact(
        k,
        tida,
        tidb,
        absorb=absorb,
        cutoff=cutoff,
        max_bond=max_bond,
        max_iterations=max_iterations,
        tol=tol,
        solver=solver,
        solver_maxiter=solver_maxiter,
        prenormalize=prenormalize,
        condition=condition,
        enforce_pos=enforce_pos,
        pos_smudge=pos_smudge,
        init=init,
        optimize=optimize,
        info=info,
        **contract_opts,
    )

    if gauges is not None:
        k.gauge_simple_remove(outer, inner)
        gauges[bix] = info["singular_values"]


def gate_cluster_(
    self: TensorNetworkGenVector,
    G,
    where,
    max_bond,
    cutoff=0.0,
    gauges=None,
    max_distance=1,
    mode="graphdistance",
    fillin=False,
    grow_from="all",
    max_iterations=100,
    tol=1e-10,
    solver="solve",
    solver_maxiter=4,
    prenormalize=False,
    condition=True,
    enforce_pos=True,
    pos_smudge=1e-10,
    init="svd",
    gauge_power=1.0,
    gauge_smudge=1e-10,
    equalize_norms=None,
    optimize="auto-hq",
    info=None,
    **contract_opts,
):
    """Apply a gate to this tensor network, using a single cluster around the
    target sites to then compress the increased bond dimension.

    Parameters
    ----------
    G : array
        The gate to apply.
    where : sequence of hashable
        The site(s) to apply the gate to.
    max_bond : int
        The maximum bond dimension to compress the bond to.
    cutoff : float, optional
        The singular value cutoff to use, once the compressed projectors
        have been fitted.
    gauges : dict[any, array], optional
        Gauges, in terms of singular value vectors, to absorb around the
        cluster.
    max_distance : int, optional
        The maximum distance to the initial tagged region, or if using
        'loopunion' mode, the maximum size of any loop.
    mode : {'graphdistance', 'loopunion'}, optional
        How to select the local tensors, either by graph distance or by
        selecting the union of all loopy regions containing ``tids``.
    fillin : bool or int, optional
        Whether to fill in the local patch with additional tensors, or not.
        `fillin` tensors are those connected by two or more bonds to the
        original local patch, the process is repeated int(fillin) times.
    grow_from : {"all", "any", "alldangle", "anydangle"}, optional
        If mode is 'loopunion', whether each loop should contain *all* of
        the initial tids, or just *any* of them (generating a larger
        region). If 'alldangle' or 'anydangle', then the individual
        loops can contain the target tids even if they are dangling.
    max_iterations : int, optional
        The maximum number of iterations to use when fitting the projectors.
    tol : float, optional
        The target tolerance to reach when fitting the projectors.
    solver : {'solve', None, str}, optional
        The solver to use inside the fitting loop. If None will use a custom
        conjugate gradient method. Else can be any of the iterative solvers
        supported by ``scipy.sparse.linalg`` such as 'gmres', 'bicgstab', etc.
    solver_maxiter : int, optional
        The maximum number of iterations to use for the *inner* solver, i.e.
        per fitting step, only for iterative `solver` args.
    prenormalize : bool, optional
        Whether to prenormalize the environment tensor such that its full
        contraction before compression is 1. Recommended for stability when
        the normalization does not matter.
    condition : bool or "iso", optional
        Whether to condition the projectors after each fitting step. If
        ``True``, their norms will be simply matched. If ``"iso"``, then they
        are gauged each time such that the previous tensor is isometric.
        Recommended for stability.
    enforce_pos : bool, optional
        Whether to enforce the environment tensor to be positive semi-definite
        by symmetrizing and clipping negative eigenvalues. Recommended for
        stability.
    pos_smudge : float, optional
        The value to clip negative eigenvalues to when enforcing positivity,
        relative to the largest eigenvalue.
    init : {'svd', 'eigh', 'random', 'reduced'}, optional
        How to initialize the compression projectors. The options are:

        - 'svd': use a truncated SVD of the environment tensor with the bra
          bond traced out.
        - 'eigh': use a similarity compression of the environment tensor with
          the bra bond traced out.
        - 'random': use random projectors.
        - 'reduced': split the environment into bra and ket parts, then
          canonize one half left and right to get the reduced factors.

    gauge_power : float, optional
        A power to raise the gauge vectors to when inserting.
    gauge_smudge : float, optional
        A small value to add to the gauge vectors to avoid singularities
        when inserting.
    equalize_norms : bool, optional
        Whether to equalize the norms of the two tensors after compression,
        stripping the exponent.
    optimize : str, optional
        Contraction path optimizer to use when forming the bond environment.
    info : dict, optional
        If provided, will store information about the fitting process here.
        The keys 'iterations' and 'distance' will contain the final number of
        iterations and distance reached respectively.
    contract_opts
        Other contraction options to pass.
    """
    if len(where) == 1:
        # simply contract single site gates in
        return self.gate_(G, where, contract=True)

    taga, tagb = map(self.site_tag, where)
    (tida,) = self._get_tids_from_tags(taga)
    (tidb,) = self._get_tids_from_tags(tagb)

    # apply gate without truncation
    if gauges is not None:
        self.gate_simple_(
            G,
            where,
            gauges=gauges,
            max_bond=None,
            cutoff=0.0,
            power=gauge_power,
            smudge=gauge_smudge,
        )
    else:
        self.gate_(
            G,
            where,
            contract="reduce-split",
            max_bond=None,
            cutoff=0.0,
        )

    compress_between_tids_bondenv_cluster(
        self,
        tida,
        tidb,
        max_bond,
        cutoff=cutoff,
        max_distance=max_distance,
        mode=mode,
        fillin=fillin,
        grow_from=grow_from,
        gauges=gauges,
        max_iterations=max_iterations,
        tol=tol,
        gauge_power=gauge_power,
        gauge_smudge=gauge_smudge,
        solver=solver,
        solver_maxiter=solver_maxiter,
        prenormalize=prenormalize,
        condition=condition,
        enforce_pos=enforce_pos,
        pos_smudge=pos_smudge,
        init=init,
        optimize=optimize,
        info=info,
        **contract_opts,
    )

    if equalize_norms:
        self.strip_exponent(tida, equalize_norms)
        self.strip_exponent(tidb, equalize_norms)


class ClusterUpdateGen(SimpleUpdateGen):
    def __init__(
        self,
        psi0,
        ham,
        max_distance=1,
        mode="graphdistance",
        fillin=0,
        grow_from="all",
        gauge_power=1.0,
        **kwargs,
    ):
        super().__init__(psi0, ham, **kwargs)

        self.gate_opts.setdefault("max_distance", max_distance)
        self.gate_opts.setdefault("mode", mode)
        self.gate_opts.setdefault("fillin", fillin)
        self.gate_opts.setdefault("grow_from", grow_from)
        self.gate_opts.setdefault("gauge_power", gauge_power)
        # we don't use the `contract` option
        self.gate_opts.pop("contract", None)

        # match gauging power in gate and equilibration by default
        self.equilibrate_opts.setdefault("power", gauge_power)

        # track compression performance
        self.max_fit_iterations = array.array("L")
        self.max_fit_distances = array.array("d")

    def presweep(self, *args, **kwargs):
        self.max_fit_iterations.append(0)
        self.max_fit_distances.append(0.0)
        return super().presweep(*args, **kwargs)

    def gate(self, U, where):
        info = {}
        gate_cluster_(
            self._psi,
            U,
            where,
            gauges=self._gauges,
            info=info,
            **self.gate_opts,
        )
        self.max_fit_iterations[-1] = max(
            info["iterations"], self.max_fit_iterations[-1]
        )
        self.max_fit_distances[-1] = max(
            info["distance"], self.max_fit_distances[-1]
        )

    def assemble_plot_data(self):
        data = super().assemble_plot_data()
        data["max_fit_iterations"] = self.max_fit_iterations
        data["max_fit_distances"] = {
            "y": self.max_fit_distances,
            "yscale": "log",
        }
        return data


def compress_between_tids_bondenv_gloop_expand(
    self: TensorNetwork,
    tida,
    tidb,
    max_bond,
    cutoff=0.0,
    gauges=None,
    gloops=None,
    autocomplete=True,
    # autoreduce=True,  # not implemented yet
    grow_from="all",
    normalized=False,
    combine="sum",
    max_iterations=100,
    tol=1e-6,
    gauge_power=1.0,
    gauge_smudge=1e-6,
    prenormalize=False,
    condition=True,
    enforce_pos=True,
    pos_smudge=1e-6,
    solver="solve",
    solver_maxiter=4,
    init="svd",
    equalize_norms=None,
    optimize="auto-hq",
    info=None,
    **contract_opts,
):
    """Compress the bond between the two tensors identified by ``tida`` and
    ``tidb`` using a generalized loop expansion approximation of the bond
    environment tensor.
    """
    from quimb.tensor.belief_propagation import RegionGraph

    ta = self.tensor_map[tida]
    tb = self.tensor_map[tidb]
    _, bix, _ = tensor_make_single_bond(ta, tb)

    if gauges is not None:
        g = gauges.pop(bix, None)
        if g is not None:
            gsqrt = do("sqrt", g)
            ta.multiply_index_diagonal_(bix, gsqrt)
            tb.multiply_index_diagonal_(bix, gsqrt)
        absorb = None
    else:
        absorb = "both"

    gloops = self.get_local_gloops(
        tids=(tida, tidb),
        gloops=gloops,
        grow_from=grow_from,
    )

    # create a region graph, for intersections and counting numbers
    rg = RegionGraph(gloops, autocomplete=autocomplete)
    E = None
    for r in rg.regions:
        # and and possibly gauge region
        k = self._select_tids(r, virtual=False)
        if gauges is not None:
            k.gauge_simple_insert(
                gauges, power=gauge_power, smudge=gauge_smudge
            )

        # contract the 4-index bond tensor given by this region
        k._cut_between_tids(tida, tidb, "kl", "kr")
        b = k.conj().reindex_({"kl": "bl", "kr": "br"})
        Er = (
            (k | b)
            .contract(
                all,
                output_inds=("kl", "kr", "bl", "br"),
                optimize=optimize,
                **contract_opts,
            )
            .data
        )

        if normalized:
            Er /= do("einsum", "aabb->", Er)

        if combine == "sum":
            # sum-combine environments, weighting by counting numbers
            # XXX: normalize by trace of Er ?
            cr = rg.get_count(r)
            if cr != 1:
                Er = cr * Er
            E = Er if E is None else E + Er

        elif combine == "prod":
            # product-combine environments, weighting by counting numbers
            cr = rg.get_count(r)
            if cr != 1:
                if cr < 0:
                    Eabs = do("abs", Er)
                    Emax = do("max", Eabs)
                    Etol = Emax * 1e-6
                    # replace zeros
                    Er = Er + Etol * (Eabs < Etol)
                Er = Er**cr
            if E is None:
                E = Er
            else:
                E = E * Er

        else:
            raise ValueError(f"Unrecognized combine method: {combine}")

    # compute the projectors for combined environment
    Pl, maybe_svals, Pr = compute_bondenv_projectors(
        E,
        max_bond=max_bond,
        cutoff=cutoff,
        max_iterations=max_iterations,
        tol=tol,
        absorb=absorb,
        solver=solver,
        solver_maxiter=solver_maxiter,
        prenormalize=prenormalize,
        condition=condition,
        enforce_pos=enforce_pos,
        pos_smudge=pos_smudge,
        init=init,
        info=info,
    )

    # absorb the projectors to compress the tn!
    ta.gate_(Pl, bix, transposed=True)
    tb.gate_(Pr, bix)

    if maybe_svals is not None:
        # update with new, truncated singular values
        gauges[bix] = maybe_svals

    if equalize_norms:
        self.strip_exponent(tida, equalize_norms)
        self.strip_exponent(tidb, equalize_norms)


def gate_gloop_expand_(
    self: TensorNetworkGenVector,
    G,
    where,
    max_bond,
    cutoff=0.0,
    gauges=None,
    gloops=None,
    autocomplete=True,
    # autoreduce=True,
    grow_from="all",
    normalized=False,
    combine="sum",
    max_iterations=100,
    tol=1e-6,
    gauge_power=1.0,
    gauge_smudge=1e-6,
    prenormalize=False,
    condition=True,
    enforce_pos=True,
    pos_smudge=1e-6,
    solver="solve",
    solver_maxiter=4,
    equalize_norms=None,
    optimize="auto-hq",
    info=None,
    **contract_opts,
):
    """Apply a gate to a pair of tensors, then compress the bond between them
    using a generalized loop expansion approximation of the bond environment
    tensor.

    Parameters
    ----------
    G : array
        The gate to apply.
    where : sequence of hashable
        The site(s) to apply the gate to.
    max_bond : int
        The maximum bond dimension to compress the bond to.
    cutoff : float, optional
        The singular value cutoff to use, once the compressed projectors
        have been fitted.
    gauges : dict[any, array], optional
        If provided, the gauge tensors to use for each bond.
    gloops : int, optional
        The maximum number of tensors to include in a single cluster.
    max_iterations : int, optional
        The maximum number of iterations to use when fitting the
        projectors.
    tol : float, optional
        The tolerance to use when fitting the projectors.
    """
    if len(where) == 1:
        # simply contract single site gates in
        return self.gate_(G, where, contract=True)

    taga, tagb = map(self.site_tag, where)
    (tida,) = self._get_tids_from_tags(taga)
    (tidb,) = self._get_tids_from_tags(tagb)

    # apply gate without truncation
    if gauges is not None:
        self.gate_simple_(
            G,
            where,
            gauges=gauges,
            max_bond=None,
            cutoff=0.0,
            power=gauge_power,
            smudge=gauge_smudge,
        )
    else:
        self.gate_(
            G,
            where,
            contract="reduce-split",
            max_bond=None,
            cutoff=0.0,
        )

    compress_between_tids_bondenv_gloop_expand(
        self,
        tida,
        tidb,
        max_bond,
        cutoff=cutoff,
        gloops=gloops,
        autocomplete=autocomplete,
        # autoreduce=autoreduce,
        grow_from=grow_from,
        normalized=normalized,
        combine=combine,
        gauges=gauges,
        max_iterations=max_iterations,
        tol=tol,
        gauge_power=gauge_power,
        gauge_smudge=gauge_smudge,
        prenormalize=prenormalize,
        condition=condition,
        enforce_pos=enforce_pos,
        pos_smudge=pos_smudge,
        solver=solver,
        solver_maxiter=solver_maxiter,
        equalize_norms=equalize_norms,
        optimize=optimize,
        info=info,
        **contract_opts,
    )


class GLoopExpandUpdateGen(SimpleUpdateGen):
    def __init__(
        self,
        psi0,
        ham,
        gloops=None,
        grow_from="all",
        normalized=False,
        combine="sum",
        enforce_pos=True,
        pos_smudge=1e-6,
        gauge_power=1.0,
        **kwargs,
    ):
        super().__init__(psi0, ham, **kwargs)

        self.gate_opts.setdefault("gloops", gloops)
        self.gate_opts.setdefault("grow_from", grow_from)
        self.gate_opts.setdefault("normalized", normalized)
        self.gate_opts.setdefault("combine", combine)
        self.gate_opts.setdefault("enforce_pos", enforce_pos)
        self.gate_opts.setdefault("pos_smudge", pos_smudge)
        self.gate_opts.setdefault("gauge_power", gauge_power)
        # we don't use the `contract` option
        self.gate_opts.pop("contract", None)

        # match gauging power in gate and equilibration by default
        self.equilibrate_opts.setdefault("power", gauge_power)

        # track compression performance
        self.max_fit_iterations = array.array("L")
        self.max_fit_distances = array.array("d")

    def presweep(self, *args, **kwargs):
        self.max_fit_iterations.append(0)
        self.max_fit_distances.append(0.0)
        return super().presweep(*args, **kwargs)

    def gate(self, U, where):
        info = {}
        gate_gloop_expand_(
            self._psi,
            U,
            where,
            gauges=self._gauges,
            info=info,
            **self.gate_opts,
        )
        self.max_fit_iterations[-1] = max(
            info["iterations"], self.max_fit_iterations[-1]
        )
        self.max_fit_distances[-1] = max(
            info["distance"], self.max_fit_distances[-1]
        )

    def assemble_plot_data(self):
        data = super().assemble_plot_data()
        data["max_fit_iterations"] = self.max_fit_iterations
        data["max_fit_distances"] = {
            "y": self.max_fit_distances,
            "yscale": "log",
        }
        return data
